"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.SimpleMQBroker = void 0;
var _events = require("events");
var _ = require("./");
function matchRule(rule, topic) {
  if (rule === topic) {
    return true;
  }
  const ruleParts = rule.split('/');
  const topicParts = topic.split('/');
  for (let i = 0; i < ruleParts.length; i++) {
    if (ruleParts[i] === '#') {
      return true;
    }
    if (ruleParts[i] !== '+' && ruleParts[i] !== topicParts[i]) {
      return false;
    }
  }
  return ruleParts.length === topicParts.length;
}
class SimpleMQBroker extends _events.EventEmitter {
  constructor() {
    let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    super();
    this.server = new _.Server();
    this.authenticate = options.authenticate;
    this.retryInterval = options.retryInterval ?? 3000;
    this.keepalive = options.keepalive ?? 60;
    this.clients = {};
    this.server.on('connection', this.handleClient.bind(this));
    this.server.on('error', this.emit.bind(this, 'error'));
    this.server.on('close', this.emit.bind(this, 'close'));
    this.server.on('listening', this.emit.bind(this, 'listening'));
    this.sessions = {};
    this.nextId = 1000;
  }
  start(port) {
    this.server.listen({
      port: port ?? 1883,
      host: '0.0.0.0'
    });
  }
  stop() {
    this.server.close();
  }
  handleClient(client) {
    let id;
    let session;
    client.on('close', () => {
      if (id && session) {
        delete this.clients[id];
        session.alive = false;
      }
    });
    client.on('disconnect', () => {
      client.end();
    });
    client.on('connect', async packet => {
      var _this$sessions$packet, _this$authenticate;
      if ((_this$sessions$packet = this.sessions[packet.clientId]) !== null && _this$sessions$packet !== void 0 && _this$sessions$packet.alive) {
        client.connack({
          returnCode: 2
        });
        return;
      }
      const version = packet.protocolVersion ?? 5;
      client.setProtocolVersion(version);
      const pass = (await ((_this$authenticate = this.authenticate) === null || _this$authenticate === void 0 ? void 0 : _this$authenticate.call(this, packet.username, packet.password))) ?? true;
      const sessionPresent = packet.clean ? false : !!this.sessions[packet.clientId];
      const keepalive = packet.keepalive ?? this.keepalive;
      if (pass) {
        client.connack({
          returnCode: 0,
          reasonCode: 0,
          sessionPresent,
          attributes: {
            serverKeepAlive: keepalive
          }
        });
        client.setKeepAlive(keepalive);
      } else {
        client.connack({
          returnCode: 5,
          reasonCode: 135
        });
      }
      id = packet.clientId;
      this.clients[id] = client;
      session = this.sessions[id] ??= {
        subs: {},
        alive: true,
        pending: {}
      };
      session.alive = true;
      if (packet.clean) {
        session.subs = {};
        session.pending = {};
        session.retry && clearTimeout(session.retry);
      }
      this.emit('connect', id);
    });
    client.on('publish', async packet => {
      if (!session) return;
      switch (packet.qos) {
        case 0:
          this.processMessage(packet.messageId, packet.topic, packet.payload);
          break;
        case 1:
          client.puback({
            messageId: packet.messageId
          });
          this.processMessage(packet.messageId, packet.topic, packet.payload);
          break;
        case 2:
          session.pending[packet.messageId] = {
            topic: packet.topic,
            payload: packet.payload
          };
          client.pubrec({
            messageId: packet.messageId
          });
          break;
      }
    });
    client.on('pubrel', packet => {
      var _session, _session2;
      const {
        messageId
      } = packet;
      if (!((_session = session) !== null && _session !== void 0 && _session.pending[messageId])) return;
      const message = session.pending[messageId];
      (_session2 = session) === null || _session2 === void 0 ? true : delete _session2.pending[messageId];
      client.pubcomp({
        messageId
      });
      this.processMessage(messageId, message.topic, message.payload);
    });
    client.on('subscribe', packet => {
      if (!session) return;
      const granted = [];
      for (const sub of packet.subscriptions) {
        session.subs[sub.topic] = sub.qos;
        granted.push(sub.qos);
      }
      client.suback({
        messageId: packet.messageId,
        granted
      });
    });
    client.on('unsubscribe', packet => {
      if (!session) return;
      for (const topic of packet.unsubscriptions) {
        var _session3;
        (_session3 = session) === null || _session3 === void 0 ? true : delete _session3.subs[topic];
      }
      client.unsuback({
        messageId: packet.messageId,
        reasonCode: 0
      });
    });
    client.on('pingreq', packet => {
      client.pingresp({
        messageId: packet.messageId
      });
    });
  }
  processMessage(messageId, topic, payload) {
    this.emit('message', topic, payload);
    this.publish(messageId, topic, payload);
  }
  publish(messageId, topic, payload) {
    const id = messageId ?? this.nextId++;
    for (const cId in this.sessions) {
      for (const rule in this.sessions[cId].subs) {
        if (matchRule(rule, topic)) {
          const qos = this.sessions[cId].subs[rule] ?? 0;
          this.doTask(cId, id, topic, payload, qos);
          break;
        }
      }
    }
  }
  doTask(cId, messageId, topic, payload, qos) {
    let dup = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : false;
    if (!this.sessions[cId]) return;
    this.clients[cId].publish({
      messageId,
      topic,
      payload,
      qos,
      dup
    });
    switch (qos) {
      case 1:
        this.clients[cId].once('puback', () => {
          var _this$sessions$cId;
          clearTimeout((_this$sessions$cId = this.sessions[cId]) === null || _this$sessions$cId === void 0 ? void 0 : _this$sessions$cId.retry);
        });
        this.sessions[cId].retry = setTimeout(() => {
          var _this$clients$cId, _this$sessions$cId2;
          (_this$clients$cId = this.clients[cId]) === null || _this$clients$cId === void 0 ? void 0 : _this$clients$cId.removeAllListeners('puback');
          (_this$sessions$cId2 = this.sessions[cId]) === null || _this$sessions$cId2 === void 0 ? true : delete _this$sessions$cId2.retry;
          this.doTask(cId, messageId, topic, payload, qos, true);
        }, this.retryInterval);
        break;
      case 2:
        this.clients[cId].once('pubrec', () => {
          this.clients[cId].pubrel();
        });
        this.clients[cId].once('pubcomp', () => {
          var _this$sessions$cId3;
          clearTimeout((_this$sessions$cId3 = this.sessions[cId]) === null || _this$sessions$cId3 === void 0 ? void 0 : _this$sessions$cId3.retry);
        });
        this.sessions[cId].retry = setTimeout(() => {
          var _this$clients$cId2, _this$clients$cId3, _this$sessions$cId4;
          (_this$clients$cId2 = this.clients[cId]) === null || _this$clients$cId2 === void 0 ? void 0 : _this$clients$cId2.removeAllListeners('pubrec');
          (_this$clients$cId3 = this.clients[cId]) === null || _this$clients$cId3 === void 0 ? void 0 : _this$clients$cId3.removeAllListeners('pubcomp');
          (_this$sessions$cId4 = this.sessions[cId]) === null || _this$sessions$cId4 === void 0 ? true : delete _this$sessions$cId4.retry;
          this.doTask(cId, messageId, topic, payload, qos, true);
        }, this.retryInterval);
        break;
    }
  }
}
exports.SimpleMQBroker = SimpleMQBroker;
var _default = SimpleMQBroker;
exports.default = _default;
//# sourceMappingURL=simple-mq.js.map